from typing import Any  # noqa
from typing import Callable, Generic, Optional, Type, TypeVar, Union, overload

from django.db import models

T = TypeVar("T")
S = TypeVar("S", bound=models.Model)
Z = TypeVar("Z", bound=Union["models.QuerySet[Any]", models.Model, None])


class ComputedRelation(Generic[T, S, Z]):
    def __init__(
        self,
        *,
        label: Optional[str] = None,
        fget: Callable[[T], Z],
        model: Union[Callable[[], Type[S]], Type[S]],
    ) -> None:
        self.name = fget.__name__

        self._model = model  # model if isinstance(model, type) else model()
        self.fget: Callable[[T], Z] = fget
        self.many_to_many = True
        self.label = label

    @property
    def related_model(self) -> Type[S]:
        return self._model if isinstance(self._model, type) else self._model()

    @overload
    def __get__(self, instance: None, own: Type[T]) -> "ComputedRelation[T, S, Z]":
        pass

    @overload
    def __get__(self, instance: T, own: Type[T]) -> Z:
        pass

    def __get__(
        self, instance: Union[T, None], own: Type[T],
    ) -> Union["ComputedRelation[T, S, Z]", Z]:
        if instance is None:
            return self

        return self.fget(instance)


def computed_relation(
    *, label: Optional[str] = None, model: Union[Callable[[], Type[S]], Type[S]],
) -> Callable[[Callable[[T], Z]], ComputedRelation[T, S, Z]]:
    def inner(fget: Callable[[T], Z]) -> ComputedRelation[T, S, Z]:
        return ComputedRelation(fget=fget, label=label, model=model)

    return inner



from typing import Any

from django.core.management.base import BaseCommand

from ... import generate_settings


class Command(BaseCommand):
    help = "Generates a JSON representation of all settings."

    def handle(self, *args: Any, **options: Any) -> None:
        self.stdout.write(generate_settings())



from typing import Any, List

import requests
import simplejson
from django.conf import settings
from django.http import HttpRequest
from django.template.defaultfilters import escape


def get_accept_list(request: HttpRequest) -> List[str]:
    """
    Given the incoming request, return a tokenized list of media
    type strings.

    From https://github.com/encode/django-rest-framework/blob/master/rest_framework/negotiation.py
    """
    header = request.META.get("HTTP_ACCEPT", "*/*")
    return [token.strip() for token in header.split(",")]


def should_respond_with_json(request: HttpRequest) -> bool:
    accepts = get_accept_list(request)

    return request.GET.get("format", None) == "json" or any(
        ["application/json" in content_type for content_type in accepts]
    )


def render_jsx_to_string(
    request: HttpRequest, template_name: str, context: Any, props: Any
) -> str:
    from reactivated import encode_complex_types

    respond_with_json = should_respond_with_json(request)

    payload = {"context": {**context, "template_name": template_name}, "props": props}
    data = simplejson.dumps(payload, indent=4, default=encode_complex_types)
    headers = {"Content-Type": "application/json"}

    if "debug" in request.GET:
        return f"<html><body><h1>Debug response</h1><pre>{escape(data)}</pre></body></html>"
    elif (
        respond_with_json or "raw" in request.GET or settings.REACTIVATED_SERVER is None
    ):
        request._is_reactivated_response = True  # type: ignore[attr-defined]
        return data

    return requests.post(  # type: ignore[no-any-return]
        f"{settings.REACTIVATED_SERVER}/__ssr/", headers=headers, data=data
    ).json()["rendered"]



from django.http import HttpRequest
from django.template.response import TemplateResponse

from . import forms


def hello_world(request: HttpRequest) -> TemplateResponse:
    return TemplateResponse(request, "hello_world.tsx", {})


def sample_form(request: HttpRequest) -> TemplateResponse:
    form = forms.SampleForm()
    return TemplateResponse(request, "sample_form.tsx", {"form": form})
